# ----------\
# cutter.pl  \
#----------------------------------------------------------------------
# Copyright (C) 2008
# See comments at bottom for GNU GPL conditions.
#----------------------------------------------------------------------

#!/usr/bin/perl

use strict;
use warnings;

use File::Spec;

# Boolean constants
my $TRUE  = 1;
my $FALSE = 0;

# Check how many arguments the user included when running the program
my $N = scalar @ARGV;

# If we don't have enough arguments, tell the user how to run the program properly and then terminate
if ( $N < 3)
{
	print "\n  USAGE:        % perl cutter.pl  <chunkSize_bp>  <shift_bp>  <lastChunk_bp >= PERCENT*chunkSize_bp \n";
	print "\n           e.g. % perl cutter.pl  10000  3000  0.50  \n";
	print "                                    means 10K chunks, start of subsequent chunks shifted 3K downstream, and \n";
	print "                                    the last chunk must be at least 50% of a chunk to be its own chunk \n";
	exit;
}

# Grab chunk size, shift size, and minimum size of last chunk from arguments
my $CHUNKSIZE_bp 		=  $ARGV[0];
my $SHIFTSIZE_bp 		=  $ARGV[1];
my $LASTCHUNKSIZE_bp 		= ($ARGV[2] * $CHUNKSIZE_bp);

# Verify that the user isn't giving us garbage.  Shift size must be > 0 and <= chunk size.
if ($SHIFTSIZE_bp < 1 or $SHIFTSIZE_bp > $CHUNKSIZE_bp)
{
	print "\n  Invalid SHIFTSIZE: [ 1 <= $SHIFTSIZE_bp <= CHUNKSIZE($CHUNKSIZE_bp) ] \n";
	exit;
}

#----------------------------------------------------------------------
# see documentation help at the bottom
#----------------------------------------------------------------------

# ==========---- PATH and FILE NAMES ---- ===========================
#my $SPLIT_DIR      = "../0_extract_from_dB/data_all_chr";      # text to split: folder containing .fna files of ASCII DNA characters
#my $SPLIT_DIR      = "../0_extract_from_dB/data_all_chr_April23_2009";      # text to split: folder containing .fna files of ASCII DNA characters
#my $SPLIT_DIR      = "../0_extract_from_dB/data_all_chr_May27_2009";      # text to split: folder containing .fna files of ASCII DNA characters
my $SPLIT_DIR      = File::Spec->catfile("..", "0_extract_from_dB", "data_all_chr");      # text to split: folder containing .fna files of ASCII DNA characters
my $OUTPUT_DIR     = "split_texts";         # new folder to put the split texts in
# ===================================================================

#----------------------------------------------------------------------
# Cross-Platform File Input/Output Code!
my @path = File::Spec->splitdir( File::Spec->rel2abs($0) );
my $abs_pathname    = File::Spec->catfile( @path[0 .. $#path - 1], "");
my $full_pathname   = $abs_pathname . $SPLIT_DIR;
my $output_pathname = $abs_pathname . $OUTPUT_DIR;

# Convert that into something that works on Windows and Linux
$full_pathname = File::Spec->abs2rel( $full_pathname, $abs_pathname );

my $a;		# The splitpath function returns three strings.  We only want one of them.
my $b;		# These two variables catch the extraneous data.
($a, $b, $output_pathname) = File::Spec->splitpath( $output_pathname );
#----------------------------------------------------------------------

my $short_title;			# Name of the bug we're using

# Tell the user we're starting, and remember when we started so we can check how long it took us
print "Processing files in directory:  $full_pathname \n\n";
my $startTime = time();                  # starts timer

# ==========================================================================
# don't run if output directory exists (avoid old results getting merged)
if (-d $output_pathname)
{
	print "ERROR:  $output_pathname already exists with old results.\n";
	print "        REMOVE $output_pathname and re-run.\n\n";
	exit(-1);
}
mkdir $output_pathname;



# grab a "glob" of files, each of the filenames is put into elements of the array @dirList
my $pathname = File::Spec->catfile( $full_pathname, '*');
my @dirList = glob($pathname);

# ==========================================================================
# process the genomes one at a time

my $n = 0;	# nth genome in progress

# for each ORGANISM
foreach my $nextORG (@dirList)
{
	#if ($n >= 28) {
		#$n++;
		#next;
	#}

	$n++;		# Increment "current genome" counter

	#print "next subdir: $nextDir \n";

        # Get the filename of this organism's FNA file
        my @pathDir = File::Spec->splitdir($nextORG);
	my $ncbi_dir_name = $pathDir[(@pathDir - 1)];


	# HERE
	#if ( $ncbi_dir_name ne "Mycoplasma_gallisepticum" )
	#{
	#	next;
	#}

        # Print organism filename
       	print "Organism filename: $ncbi_dir_name \n";

	# glob all the files
	my $pathDir = File::Spec->catfile($nextORG, "*fna");
	my @files_to_split = glob ($pathDir);

        #print "pathDir: $pathDir \n";

	my $fileNOfna;			       			# Stores chromosome ID number

	foreach my $nextFILE (@files_to_split)
	{
		 # open the file to read
		 open (my $filename, '<', $nextFILE) or die "Cannot open $nextFILE !";

		 my $output_filename;  				# stores the new filename for each text

                 #print "Next filename: $nextFILE \n";

		 # get the short title of the bug
                 my @short_path = File::Spec->splitdir( $nextFILE );
                 $short_title = File::Spec->catfile( @short_path[1 .. $#short_path] );

                 #print "Short title: $short_title \n";

		 # determine which chromosome this is (snag from the end of the filename)
		 $nextFILE =~ m/_(chr..)\.fna/;
		 $fileNOfna = $1;

                 #print "File #: $fileNOfna \n";


		 # grab the DNA (handle multiple lines)
		 my $totalText = "";				# Stores the entire DNA sequence

                 # Read in DNA, one line at a time.
		 while ( my $line = <$filename> )
		 {
                 	  # Remove formatting and concatenate onto string holding the entire DNA sequence
			  chomp $line;
			  $totalText = $totalText . $line;
                          #print "Line: $line \n";            	# Test - comment out later
		 }


		 # time to cut ....
		 my $endBP;					# Nucleotide the chunk ends on
		 my $startBP = 0;                               # Nucleotide the chunk begins on
		 my $totalLengthBP = length($totalText);	# Total length of the entire DNA sequence

		 my $done = $FALSE;				# Sentinel to check if we're done or not
		 my $chunkNumber = 0;				# Keeps track of how many chunks we have so far

                 # Keep going until we're done cuttin or we're out of data
		 while ( !$done  &&  ($startBP < $totalLengthBP) )
		 {
			$chunkNumber++;				# Increment chunk number counter

                        # Grab the next chunk
			my $chunk = substr($totalText, $startBP, $CHUNKSIZE_bp);

			# if this chuck is a full one, what about the next one?
			if ( length($chunk) == $CHUNKSIZE_bp )
			{
                        	# Calculate end position of this chunk
				$endBP = $startBP + $CHUNKSIZE_bp - 1;

				# is next chunk big enough to be its own chunk?
				my $remainingBP = $totalLengthBP - ($startBP + $SHIFTSIZE_bp);
				if ( $remainingBP < $LASTCHUNKSIZE_bp )		# If final chunk is too small
				{
                               		# Grab everything from the start of the second-to-last chunk to the end of the sequence
					# redo last substr to get this tailing end of the genome (not big enough alone to be a chunk)
					$chunk = substr($totalText, $startBP);
					$endBP = $totalLengthBP - 1;			# End of this chunk == end of sequence
					$done = $TRUE;					# We're done.
                                        #print "\tLast chunk was NOT big enough in Test $fileNOfna.\n";
				}

			}

                        # Final chunk is big enough to qualify as its own chunk.  Don't merge it with the previous chunk.
			else # that last chunk was not a full chunk, so mark that as the last one
			{
				$endBP = $totalLengthBP - 1;				#  Endpoint == end of sequence

				$done = $TRUE;						# We're done.
			}


			# echo current chunk range (Comment this test stuff out later)
			#print "$fileNOfna: [$startBP, $endBP]\n";
                        #print "Chunk: $chunk \n";

			# +++++++++++++ make directories and output chunk +++++++++++++++++
  			my $newDir	= File::Spec->catdir( $output_pathname, "data_all_chr" );
  			mkdir $newDir unless (-d $newDir);	# Only make the directory if a directory by this name doesn't exist

  			my $bugNamePlusChrNum = $ncbi_dir_name."__".$fileNOfna;
  			$newDir	= File::Spec->catdir( $newDir, $bugNamePlusChrNum );
  			mkdir $newDir unless (-d $newDir);      # Only make the directory if a directory by this name doesn't exist

                        # Include the chunk number in the filename when outputting results
  			my $chunk_filename = sprintf("c%03d_", $chunkNumber);

  			#$output_filename = $chunk_filename . $ncbi_dir_name . ".fna";
  			# or alternatively, using start-end bp in name
  			my $output_filename = $chunk_filename . $ncbi_dir_name."_".$fileNOfna."_".$startBP."_to_".$endBP.".fna";

			my $outputFilename	= File::Spec->catfile( $newDir, $output_filename );

			# Attempt to create the output file (terminate on failure)
			open (my $CHUNKFILE, '>', $outputFilename) or die "Died making $outputFilename";

                        # Print results (current chunk) to file
			print $CHUNKFILE "$chunk";
			close $CHUNKFILE;

			# Are we done yet?
			if ( !$done ) # prepare for next chunk
			{
				$startBP += $SHIFTSIZE_bp;
			}

		 } # while more chunks to get

	} # FOREACH file in folder

} # for each ORGANISM



# ==========================================================================

# We're done!  Check what time it is and use this to calculate how long it took.
my $endTime = time();
my $totalMins = int ($endTime - $startTime) / 60;

# Print results
printf "\n\nTotal time for cutting bugs was %d minutes.\n\n", $totalMins;
print "\n\nDone.\n";

# End code.  Documentation below.
# ==========================================================================

__END__

=head1 NAME

C<cutter.pl>

=head1 SYNOPSIS

% perl cutter.pl  <Lmer>  <#bp per chunk>  <#bp overlap between chunks >  <last chunk must be this % chunk size>
e.g.
% perl cutter.pl  4  10000  3000  0.5

means chunks of 10000bp, each subsequent chunk will start 3000bp downstream,
and ensure that the last chuck is at least 50% of the present chunk size;
this means the last chunk(file) is in the size range of:
[ 50% of chunk size .. 1.5x chunk size)

=head1 DESCRIPTION

=head2 SUMMARY

 Cutter.pl ("Script #1") is the second of a suite of scripts designed to assist in the analysis of DNA.  This particular script breaks a large DNA sequece down into several smaller chunks of user-determined size.  The script then shifts downstream by a user-determined amount (note that we recommend chunk size and shift size should be the same) and repeats the process until it reaches the end of the sequence.  If the last chunk is large enough in proportion to the previous chunk (again, this is determined by the user) it will be treated as its own separate chunk; otherwise it will be added on to the previous chunk.  This process can be repeated multiple times for several files as needed.

 The "chunks" produced by this script can be passed to motifCounts.pl ("Script #2") for analysis.

=head2 INPUT

 Run Cutter.pl using FNA files extracted from your database using the script extract_ALL_chr.pl.

=head2 OUTPUT

 Cutter.pl produces several FNA files containing discrete "chunks" of DNA for use with  Script #2 (motifCounts.pl).  These smaller chunks are much more manageable (and therefore more preferable) than using entire genomes at a time.

=head1 AUTHORS

 Christina Nelson
 Amos C. Jones
 Mark LeBlanc

=head1 MODIFICATION HISTORY

=head3 June 1, 2010 (nkf) --

 Fixed some potential problems with regex across platforms.

=head3 May 31, 2010 (nkf) --

 Improved the internal documentation.  Wrote some of the readme documentation.
 Fixed a bug that caused "tailing" chunks to get deleted instead of added to the end of the previous chunk.
 Modified file I/O code to implement cross-platform functionality.

=head3 Jan 08, 2009 (mdl) --

 added separate output directory for each chromosome
 for example:  in ~split_texts/data_all_chr/
 Agrobacterium_tumefaciens_C58_Cereon__chr01/
 Agrobacterium_tumefaciens_C58_Cereon__chr02/

=head3 Nov 18, 2008 (mdl) --

 changed to bp rather than #words

=head3 October 01, 2008 (cn) --

 reworked for DNA

=head3 June 17, 2008 (acj) --

 changed output file names to include genre and manuscript and text numbers

 reworked algorithm to work work with chunk sizes and shift sizes that are
 not multiples of each other

=head3 June 16, 2008 (acj) --

 finished commenting and wrote pod

 fixed the last chunk problem

=head3 June 12, 2008 (acj) --

 wrote and debugged script (for anglo-saxon files =:o  )










=head1 COPYRIGHT INFORMATION


 ==========================================================================
   Copyright (C) 2009  Wheaton Genomics Research Group, Norton MA

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.